# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Pedro Algarvio (pedro@algarvio.me)`
    :copyright: Copyright 2015 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    pytestsalt.version
    ~~~~~~~~~~~~~~~~~~

    pytest salt plugin version information
'''

# Import Python Libs
from __future__ import absolute_import


__version_info__ = (2017, 8, 23)
__version__ = '{0}.{1}.{2}'.format(*__version_info__)
